#include <ch32v00x.h>
#include <stdbool.h>
#include <debug.h>

volatile bool dataReady;
volatile int encoderDir;
#define DIR_COUNTERCLOCKWISE 0
#define DIR_CLOCKWISE 1

__attribute__((interrupt("WCH-Interrupt-fast")))
void TIM2_IRQHandler() {
	if (TIM_GetITStatus(TIM2, TIM_IT_Update) != RESET) {
		encoderDir = (TIM2->CTLR1 & (1 << 4)) >> 4;
		dataReady = true;
		TIM_ClearITPendingBit(TIM2, TIM_IT_Update);
	}
}

void configurePWMChannels(int dutyCycle) {
	// TIM_OCMode = 0x40 force la sortie PWM à 0.
	// TIM_OCMode = 0x50 force la sortie PWM à 1.
	// Voir section "10.4.7 Compare/Capture Control Register 1 (TIM1_CHCTLR1)"
	// p. 104 du manuel de référence du CH32V003 (valeurs 100 et 101 de OC2M[2:0]).
	TIM_OCInitTypeDef ocInit = { 0 };
	ocInit.TIM_OCMode = (dutyCycle <= 0) ? 0x40 : ((dutyCycle < 100) ? TIM_OCMode_PWM1 : 0x50);
	ocInit.TIM_OCPolarity = TIM_OCPolarity_High;
	ocInit.TIM_OutputState = TIM_OutputState_Enable;
	ocInit.TIM_Pulse = (dutyCycle < 0) ? -dutyCycle : dutyCycle;
	TIM_OC3Init(TIM1, &ocInit);
	ocInit.TIM_OCMode = (dutyCycle >= 0) ? 0x40 : ((dutyCycle > -100) ? TIM_OCMode_PWM1 : 0x50);
	TIM_OC4Init(TIM1, &ocInit);
}

#define DUTY_CYCLE_INCREMENT 5

int main() {
	USART_Printf_Init(115200);

	int dutyCycle = 0;

	// -- On configure les entrées du timer 2.
	// PD4 = T2CH1 = Encodeur CLK
	// PD3 = T2CH2 = Encodeur DT
	RCC_APB2PeriphClockCmd(RCC_APB2Periph_GPIOD, ENABLE);
	GPIO_InitTypeDef gpioInit = { 0 };
	gpioInit.GPIO_Pin = GPIO_Pin_3 | GPIO_Pin_4;
	gpioInit.GPIO_Mode = GPIO_Mode_IN_FLOATING;
	GPIO_Init(GPIOD, &gpioInit);

	// -- On configure les sorties du timer 1.
	// PC3 = T1CH3 = L9110S IA
	// PC4 = T1CH4 = L9110S IB
	RCC_APB2PeriphClockCmd(RCC_APB2Periph_GPIOC, ENABLE);
	gpioInit.GPIO_Pin = GPIO_Pin_3 | GPIO_Pin_4;
	gpioInit.GPIO_Mode = GPIO_Mode_AF_PP;
	gpioInit.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_Init(GPIOC, &gpioInit);

	// On active le timer 1.
	RCC_APB2PeriphClockCmd(RCC_APB2Periph_TIM1, ENABLE);

	// On configure la base de temps du timer 1.
	TIM_TimeBaseInitTypeDef timeBaseInit = { 0 };
	// On veut une fréquence du signal de sortie de 20Hz, donc il faut diviser
	// l'horloge principale par 48000000 / 100 (compteur) / 20 (Hz) = 24000
	timeBaseInit.TIM_Prescaler = 24000 - 1;
	// On veut pouvoir indiquer le rapport cyclique directement en %,
	// donc le compteur devra compter de 0 à 99.
	timeBaseInit.TIM_Period = 100 - 1;
	timeBaseInit.TIM_ClockDivision = TIM_CKD_DIV1;
	timeBaseInit.TIM_CounterMode = TIM_CounterMode_Up;
	TIM_TimeBaseInit(TIM1, &timeBaseInit);

	// On configure T1CH3 et T1CH4 en fonction du rapport cyclique initial.
	configurePWMChannels(dutyCycle);

	// On active les sorties PWM.
	TIM_CtrlPWMOutputs(TIM1, ENABLE);

	// On démarre le compteur.
	TIM_Cmd(TIM1, ENABLE);

	// On active le timer 2.
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM2, ENABLE);

	// On configure la base de temps du timer 2.
	timeBaseInit.TIM_Period = 1;
	timeBaseInit.TIM_Prescaler = 0;
	// On met CKD à 2, donc la fréquence d'échantillonage des filtres est 48MHz / 4 = 12MHz.
	timeBaseInit.TIM_ClockDivision = TIM_CKD_DIV4;
	timeBaseInit.TIM_CounterMode = TIM_CounterMode_Up;
	TIM_TimeBaseInit(TIM2, &timeBaseInit);

	// On configure le timer en mode encodeur.
	TIM_EncoderInterfaceConfig(TIM2, TIM_EncoderMode_TI2, TIM_ICPolarity_Falling, TIM_ICPolarity_Falling);

	// On utilise le filtre d'entrée pour éliminer les rebonds.
	TIM_ICInitTypeDef icInit;
	icInit.TIM_Channel = TIM_Channel_1;
	icInit.TIM_ICPolarity = TIM_ICPolarity_Falling;
	icInit.TIM_ICSelection = TIM_ICSelection_DirectTI;
	icInit.TIM_ICPrescaler = TIM_ICPSC_DIV1;
	icInit.TIM_ICFilter = 12;
	TIM_ICInit(TIM2, &icInit);

	// On active les interruptions de mise à jour.
	NVIC_InitTypeDef nvicInit = { 0 };
	nvicInit.NVIC_IRQChannel = TIM2_IRQn;
	nvicInit.NVIC_IRQChannelPreemptionPriority = 0;
	nvicInit.NVIC_IRQChannelSubPriority = 1;
	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	NVIC_Init(&nvicInit);
	TIM_ITConfig(TIM2, TIM_IT_Update, ENABLE);

	// On démarre le compteur.
	TIM_Cmd(TIM2, ENABLE);

	dataReady = false;

	while (1) {
		if (dataReady) {
			dataReady = false;
			switch (encoderDir) {
			case DIR_COUNTERCLOCKWISE:
				if (dutyCycle >= (-100 + DUTY_CYCLE_INCREMENT)) {
					dutyCycle -= DUTY_CYCLE_INCREMENT;
				}
				break;

			case DIR_CLOCKWISE:
				if (dutyCycle <= (100 - DUTY_CYCLE_INCREMENT)) {
					dutyCycle += DUTY_CYCLE_INCREMENT;
				}
				break;
			}

			configurePWMChannels(dutyCycle);
			printf("Nouveau rapport cyclique : %d%%\r\n", dutyCycle);
		}
	}
}
